from safetydance_test import TestStepPrefix
from safetydance_test.step_extension import all_steps_as_step_extensions_from
from type_extensions import extension_property
from . import steps


class Http:
    '''
    Context and steps for interacting with HTTP web services.
    '''
    ...


_HTTP = Http()

@extension_property
def http(self: TestStepPrefix) -> Http:
    return _HTTP


all_steps_as_step_extensions_from(steps, target_type=Http)



import os
import sys


def pytest_addoption(parser):
    parser.addoption(
        "--no-pkgroot",
        action="store_true",
        default=False,
        help="Remove package root directory from sys.path, ensuring that "
        "rest_framework is imported from the installed site-packages. "
        "Used for testing the distribution.",
    )


def pytest_configure(config):
    from django.conf import settings

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={
            "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}
        },
        SITE_ID=1,
        SECRET_KEY="not very secret in tests",
        USE_I18N=True,
        USE_L10N=True,
        STATIC_URL="/static/",
        ROOT_URLCONF="example.urls",
        TEMPLATE_LOADERS=(
            'django.template.loaders.filesystem.Loader',
            'django.template.loaders.app_directories.Loader',
        ),
        MIDDLEWARE=(
            "django.middleware.common.CommonMiddleware",
            "django.contrib.sessions.middleware.SessionMiddleware",
            'django.middleware.csrf.CsrfViewMiddleware',
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "django.contrib.messages.middleware.MessageMiddleware",
        ),
        INSTALLED_APPS=(
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.sites",
            'django.contrib.messages',
            "django.contrib.staticfiles",

            "rest_framework",
            "rest_framework.authtoken",

            "example",
        ),
        PASSWORD_HASHERS=(
            'django.contrib.auth.hashers.SHA1PasswordHasher',
            'django.contrib.auth.hashers.PBKDF2PasswordHasher',
            'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
            'django.contrib.auth.hashers.BCryptPasswordHasher',
            'django.contrib.auth.hashers.MD5PasswordHasher',
            'django.contrib.auth.hashers.CryptPasswordHasher',
        ),
        ALLOWED_HOSTS=[
            "testserver"
        ],
    )

    try:
        import oauth_provider  # noqa
        import oauth2  # noqa
    except ImportError:
        pass
    else:
        settings.INSTALLED_APPS += (
            'oauth_provider',
        )

    try:
        import provider  # noqa
    except ImportError:
        pass
    else:
        settings.INSTALLED_APPS += (
            'provider',
            'provider.oauth2',
        )

    # guardian is optional
    try:
        import guardian  # noqa
    except ImportError:
        pass
    else:
        settings.ANONYMOUS_USER_ID = -1
        settings.AUTHENTICATION_BACKENDS = (
            'django.contrib.auth.backends.ModelBackend',
            'guardian.backends.ObjectPermissionBackend',
        )
        settings.INSTALLED_APPS += (
            'guardian',
        )

    if config.getoption("--no-pkgroot"):
        sys.path.pop(0)

        # import scoped_rbac before pytest re-adds the package root directory.
        import scoped_rbac # noqa

        package_dir = os.path.join(os.getcwd(), "scoped_rbac")
        assert not scoped_rbac.__file__.startswith(package_dir)

    try:
        import django # noqa
        django.setup()
    except AttributeError:
        pass



from django.contrib.auth.models import User
from .models import Simple
from rest_framework.viewsets import ModelViewSet
from .serializers import (
    SimpleSerializer,
    UserSerializer,
)


class SimpleViewSet(ModelViewSet):
    """
    TODO restrict queryset to authorized contexts for user
    """

    queryset = Simple.objects.all()
    serializer_class = SimpleSerializer


class UserViewSet(ModelViewSet):
    """
    API which allows users to be viewed or edited.
    TODO restrict queryset to authorized contexts for user
    """

    queryset = User.objects.all()
    serializer_class = UserSerializer



from django.urls import reverse
from safetydance_django.steps import *
from safetydance_test import scripted_test
import pytest


@pytest.mark.django_db
@scripted_test
def test_get():
    get(reverse('simple-list'))
    status_code_is(200)



from django.contrib.auth.models import User
from rest_framework import serializers
from .models import Simple


class SimpleSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Simple


class UserSerializer(serializers.HyperlinkedModelSerializer):
    """User Serialization"""

    class Meta:
        model = User
        fields = ("url", "username", "email", "first_name", "last_name", "id")



from django.conf.urls import include, url
from rest_framework.routers import DefaultRouter
from rest_framework.schemas import get_schema_view
from . import rest

router = DefaultRouter()
router.register(r"simple", rest.SimpleViewSet)
router.register(r"users", rest.UserViewSet)
schema_view = get_schema_view("Test safetydance_django API")

urlpatterns = [
    url(r"^api/", include(router.urls), name="api"),
    url(r"^api/schema", schema_view, name="api-schema"),
    url(
        r"^api/rest-framework/auth",
        include("rest_framework.urls", namespace="rest_framework"),
    ),
]



from django.db import models


class ModelForTesting(models.Model):
    """Mixin class for models used in testing."""

    class Meta:
        app_label = "example"
        abstract = True


class Simple(ModelForTesting):
    name = models.CharField(max_length=128)



from collections import OrderedDict
from copy import deepcopy
from safetydance import step, step_data
from rest_framework.test import APIClient
import pytest


http_client = step_data(APIClient, initializer=APIClient)
http_response = step_data(None)


@step
def defaults(**kwargs):
    http_client.defaults = {**http_client.defaults, **kwargs}

@step
def force_authenticate(*args, **kwargs):
    """Can be used by rest_framework.test.APIClient"""
    http_client.force_authenticate(*args, **kwargs)

@step
def force_login(*args, **kwargs):
    """Can be used by django.test.Client"""
    http_client.force_login(*args, **kwargs)

@step
def login(*args, **kwargs):
    http_client.login(*args, **kwargs)

@step
def delete(*args, **kwargs):
    '''Perform HTTP DELETE'''
    http_response = http_client.delete(*args, **kwargs)

@step
def get(*args, **kwargs):
    '''Perform HTTP GET'''
    http_response = http_client.get(*args, **kwargs)

@step
def post(*args, **kwargs):
    http_response = http_client.post(*args, **kwargs)

@step
def put(*args, **kwargs):
    http_response = http_client.put(*args, **kwargs)

@step
def status_code_is(expected):
    '''
    Check that the expected status code matches the received status code
    '''
    assert http_response.status_code == expected

@step
def content_type_is(expected):
    assert 'Content-Type' in http_response
    assert http_response['Content-Type'].startswith(expected)

@step
def response_json_is(expected):
    '''
    Check that the expected response json body matches the received response
    body.
    '''
    self.content_type_is('application/json')
    observed = http_response.json()
    assert json_values_match(expected, observed)

@step
def response_data_is(expected, excluded_fields=None):
    '''
    Check that the expected response json body matches the received response
    body.
    '''
    def clean_item(obj):
        obj.pop('id', None)
        obj.pop('url', None)
        if excluded_fields is not None and isinstance(excluded_fields, list):
            for k in excluded_fields:
                obj.pop(k, None)
        return obj

    self.content_type_is('application/json')

    observed = clean_item(deepcopy(http_response.data))
    expected = clean_item(deepcopy(expected))
    assert expected == observed

@step
def assert_data(expected, observed, excluded_fields=None):
    '''
    Check that the expected response json body matches the received response
    body.
    '''
    def clean_item(obj):
        obj.pop('id', None)
        obj.pop('url', None)
        if excluded_fields is not None and isinstance(excluded_fields, list):
            for k in excluded_fields:
                obj.pop(k, None)
        return obj

    observed = clean_item(deepcopy(observed))
    expected = clean_item(deepcopy(expected))
    assert expected == observed

@step
def response_data_list_is(list_expected, excluded_fields=None):
    '''
    Check that the expected response json body matches the received response
    body.
    '''
    def clean_item(obj):
        obj.pop('id', None)
        obj.pop('url', None)
        if excluded_fields is not None and isinstance(excluded_fields, list):
            for k in excluded_fields:
                obj.pop(k, None)
        return obj
    self.content_type_is('application/json')

    _list_observed = deepcopy(http_response.data)
    list_expected = deepcopy(list_expected)

    list_observed = []

    is_enveloped = None

    for item in _list_observed:
        if is_enveloped is None:
            if 'etag' in item:
                is_enveloped = True
            else:
                is_enveloped = False
        if is_enveloped:
            item = item['content']
        elif isinstance(item, OrderedDict):
            item = dict(item)

        list_observed.append(clean_item(item))

    for i, item in enumerate(list_expected):
        list_expected[i] = clean_item(item)

    assert list_expected == list_observed

@step
def response_url_is(url_expected):
    assert http_response.url == url_expected

@step
def response_location_header_is(location_expected):
    header = http_response['location']
    assert header == location_expected


def json_values_match(expected, observed):
    '''
    Recursively walk the expected json value validating that the observed json
    value matches. Returns False on any missing or mismatched value.
    '''
    if isinstance(expected, list):
        return lists_match(expected, observed)
    if isinstance(expected, dict):
        return dictionaries_match(expected, observed)
    return expected == observed


def lists_match(expected, observed):
    '''
    Recursively walk the expected list validating that the observed list
    matches. Returns False on any missing or mismatched value.
    '''
    if not isinstance(observed, list):
        return False
    if len(expected) != len(observed):
        return False
    for i in range(0, len(expected)):
        if not json_values_match(expected[i], observed[i]):
            return False
    return True


def dictionaries_match(expected, observed):
    '''
    Recursively walk the expected dictionary validating that the observed
    dictionary matches. Returns False on any missing key or mismatched
    value.
    Note that keys that occur in observed or its nested dictionary objects
    that don't occur in expected or its matching nested dictionary objects
    are ignored.
    '''
    if observed is None:
        return False
    for key, value in expected.items():
        if not json_values_match(value, observed.get(key)):
            return False
    return True



from django.urls import reverse
from safetydance_django.test import *
from safetydance_test import scripted_test, Given, When, Then, And
import pytest

@pytest.mark.django_db
@scripted_test
def test_get():
    Given.http.get(reverse('simple-list'))
    Then.http.status_code_is(200)
